/* Copyright (C) 2002, by Sandor Fazekas and Janos Torok */
/* Last modified: $Date: 2005/01/25 18:41:57 $ $Revision: 1.37 $ */

#ifndef BODY_H
#define BODY_H

#define BT_EOL		0
#define BT_Wall		1
#define BT_Ball		2
#define BT_Cylinder	3
#define BT_Disk		4
#define BT_BallCC	5
#define BT_Force	6
#define BT_STBall	7
#define BT_DTEdge	8
#define BT_DGBall	9
#define BT_Ellipsoid	10
#define BT_Arrow	12
#define BT_Cone		13
#define BT_Triangle	14
#define BT_Triangle2	15
#define BT_NULL		16
#define BT_N_BODY	17

/* body parameter (type 4) */
#define BP_Mask		31
#define BP_Pos		32
#define BP_Orient	64
#define BP_Coord	128
#define BP_Velo		256
#define BP_Force	512
#define BP_AngVelo	1024
#define BP_Torque	2048
#define BP_TrVelo	4096
#define BP_TrForce	8192
#define BP_TrAngVelo	16384
#define BP_Handle	32768
#define BP_LifeTime	65536
#define BP_MatPar	131072
#define BP_TrCoord	262144
#define BP_TrTorque	524288
#define BP_Dvx		1048576
#define BP_Dvy		2097152
#define BP_Dvz		4194304

#define BP_Pos_I	2
#define BP_Orient_I	3
#define BP_Coord_I	4
#define BP_Velo_I	5
#define BP_Force_I	6
#define BP_AngVelo_I	7
#define BP_Torque_I	8
#define BP_TrVelo_I	9
#define BP_TrForce_I	10
#define BP_TrAngVelo_I	11
#define BP_Handle_I	12
#define BP_LifeTime_I	13
#define BP_MatPar_I	14
#define BP_TrCoord_I	15
#define BP_TrTorque_I	16
#define BP_Dvx_I	17
#define BP_Dvy_I	18
#define BP_Dvz_I	19


#define BA_Texture1D	1
#define BA_Texture2D	2
#define BA_Color	4
#define BA_Blend	8

#define CG_Load		14

#define CEF_PARNUM	21	/* number of parameter types */
#define CEF_SUBPARNUM	5	/* number of subparameters */
#define RPN_StrLen	10	/* String length of function arrays */
#define BG_RPN_globalvarnum 6

#define OBJMAX 100

/* objects */
typedef struct BODY_ {
  int type;		/* type flag  BT_... */
  int obj;		/* object group it belongs to */
  int par;		/* parameters present BP_... */
  double x,y,z;		/* position */
  double rx,ry,rz;	/* rotation direction */
  double phi;		/* rotation angle in degrees */
  double sx,sy,sz;	/* size */
  int appearance;	/* texture, color, material BA_... */
  float color[4];
  int tex;		/* texture */
  int blend;		/* transparency */
  int ind1,ind2;	/* force index parameter */
  double params[CEF_PARNUM*CEF_SUBPARNUM];	/* parameters */
} BODY;

BODY *body;

typedef struct RestBODY_ {
  int show;		/* if to be shown */
  int type;		/* type flag  BT_... */
  int blend;		/* parameters present BP_... */
  double x,y,z;		/* position */
  double rx,ry,rz;	/* rotation diretciont */
  double phi;		/* rotation angle in degrees */
  double sx,sy,sz;	/* size */
  int tex;		/* texture */
  float color[4];
} RestBODY;

RestBODY *restb;

typedef struct bodydisplaystr_ {
  int show;		/* 1-> show this type of object */
  int blend;		/* 1-> this object is transparent */
  int texture;		/* texture number+1, 0-> no texture */
  int type;		/* type of object */
  int par;		/* parameters of the object */
  int atom;		/* Atom type in xyz format */
  float color[4];	/* color if no texture */
} bodydisplaystr;

typedef struct bodyglobal_ {
  int body_num;			/* actual number of bodies */
  int wall_num;			/* actual number of bodies */
  int input_type;		/* input file structure */
  int line_length;		/* line length */
  int active_obj;		/* subframe's object */
  int obj_num;			/* number of different objects */
  int cg_pos;			/* position of the first cg object */
  bodydisplaystr *param;	/* show object type */
  int *gbeg;			/* num of first object in group */
  int *gnum;			/* number of objects in group */
  int cgnum;			/* number of constraints */
  int new_num;			/* number of created bodies+body_num */
  int save_new;                 /* save new objects */
  char rpnglobstr[BG_RPN_globalvarnum][RPN_StrLen];/* the string of them */
  double *rpnglobpoint[BG_RPN_globalvarnum];	/* the data */
} bodyglobal;

bodyglobal bg;

BODY lb;

#define CG_NP 7

typedef struct constrglobal_ {
  int type;
  int arrayvar[CG_NP];
  RPN *p[CG_NP];
  char *text;
  int textpnum;
  RPN **textparam;
} constrglobal;

constrglobal *cg;

typedef struct constrvarglobal_ {
  vect *var;
  int *vc;
  int var_num;
} constrvarglobal;

constrvarglobal cvg;
#define RPN_var

extern const char cef_relstr[6][2];
extern const char cef_parmainstr[CEF_PARNUM][10];

#endif
