#include <ft2build.h>
#include <freetype/freetype.h>
#include <freetype/ftglyph.h>
#include <freetype/ftoutln.h>
#include <freetype/fttrigon.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <stdarg.h>
#include <stdio.h>
#include <math.h>
#include "font.h"
#include "global.h"
#include "rpn.h"
#include "body.h"

/* Prototypes */
void perr(char *, ...);
int RPN_decode(char *,RPN **,char **,int,int *,int,char [][RPN_StrLen]);
double RPN_Evaluate(RPN *,double *,double **,double *,vect *,int *,int,int);

inline int next_p2 ( int a )
{
  int rval=1;
  while(rval<a) rval<<=1;
  return rval;
}


void make_dlist(FT_Face face,int ch,GLuint list_base,GLuint *tex_base)
{
  FT_Glyph glyph;
  FT_BitmapGlyph bitmap_glyph;
  FT_Bitmap bitmap;
  int width,height,i,j;
  GLubyte *expanded_data;
  double x,y;

  if (FT_Load_Glyph(face,FT_Get_Char_Index(face,ch),FT_LOAD_DEFAULT)) {
    perr("FT_Load_Glyph failed\n");
  }
  if (FT_Get_Glyph(face->glyph,&glyph)) perr("FT_Get_Glyph failed\n");
  FT_Glyph_To_Bitmap(&glyph,ft_render_mode_normal,0,1);
  bitmap_glyph=(FT_BitmapGlyph) glyph;
  bitmap=bitmap_glyph->bitmap;
  width=next_p2(bitmap.width);
  height=next_p2(bitmap.rows);
  expanded_data=(GLubyte *)malloc(2*width*height*sizeof(GLubyte));
  for (j=0; j <height;j++) {
    for (i=0; i < width; i++) {
      expanded_data[2*(i+j*width)]= expanded_data[2*(i+j*width)+1] =
        (i>=bitmap.width || j>=bitmap.rows) ?
        0 : bitmap.buffer[i + bitmap.width*j];
    }
  }
  glBindTexture( GL_TEXTURE_2D, tex_base[ch]);
  glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, width, height,
      0, GL_LUMINANCE_ALPHA, GL_UNSIGNED_BYTE, expanded_data );
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
  free(expanded_data);

  glNewList(font.list_base+ch,GL_COMPILE);
  glBindTexture(GL_TEXTURE_2D,tex_base[ch]);
  glTranslatef(bitmap_glyph->left,0,0);
  glPushMatrix();
  glTranslatef(0,bitmap_glyph->top-bitmap.rows,0);
  x=(double)bitmap.width / (double)width,
  y=(double)bitmap.rows / (double)height;
  glBegin(GL_QUADS);
  glTexCoord2d(0,0); glVertex2f(0,bitmap.rows);
  glTexCoord2d(0,y); glVertex2f(0,0);
  glTexCoord2d(x,y); glVertex2f(bitmap.width,0);
  glTexCoord2d(x,0); glVertex2f(bitmap.width,bitmap.rows);
  glEnd();
  glPopMatrix();
  glTranslatef(face->glyph->advance.x >> 6 ,0,0);
  glEndList();

  glNewList(font.my_list_base+ch,GL_COMPILE);
  glBindTexture(GL_TEXTURE_2D,tex_base[ch]);
  glEnable(GL_TEXTURE_2D);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glEnable(GL_BLEND);
  glTranslatef(bitmap_glyph->left*0.005,0,0);
  glTranslatef(0,(bitmap_glyph->top-bitmap.rows)*0.005,0);
  x=(double)bitmap.width / (double)width,
  y=(double)bitmap.rows / (double)height;
  glBegin(GL_QUADS);
  glNormal3f(0.0,0.0,1.0);
  glTexCoord2d(0,0); glVertex2f(0,bitmap.rows*0.005);
  glTexCoord2d(0,y); glVertex2f(0,0);
  glTexCoord2d(x,y); glVertex2f(bitmap.width*0.005,0);
  glTexCoord2d(x,0); glVertex2f(bitmap.width*0.005,bitmap.rows*0.005);
  glEnd();
  glTranslatef((face->glyph->advance.x >> 6)*0.005 ,0,0);
  glTranslatef(0,-(bitmap_glyph->top-bitmap.rows)*0.005,0);
  glEndList();
}



void font_init(const char * fname, unsigned int h,int set)
{
  FT_Library library;
  FT_Face face;
  int i;

  if (set)  {
    font.font=fname;
  } else {
    if (font.font==NULL) font.font=fname;
  }
  font.h=(double)h;
  if (FT_Init_FreeType(&library)) perr("FT_Init_FreeType failed\n");
  if (FT_New_Face(library,font.font,0,&face))  {
    if (FT_New_Face(library,"VeraBd.ttf",0,&face))  {
      perr("FT_New_Face failed (probably a problem with your font file)\n");
    }
  }
  FT_Set_Char_Size(face,h << 6,h << 6,96,96);
  font.list_base=glGenLists(128);
  font.my_list_base=glGenLists(128);
  glGenTextures(128,font.textures);
  for(i=0;i<128;i++) {
    make_dlist(face,i,font.list_base,font.textures);
  }
  FT_Done_Face(face);
  FT_Done_FreeType(library);
}

inline void pushScreenCoordinateMatrix()
{
  GLint  viewport[4];

  glPushAttrib(GL_TRANSFORM_BIT);
  glGetIntegerv(GL_VIEWPORT, viewport);
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  gluOrtho2D(viewport[0],viewport[2],viewport[1],viewport[3]);
  glPopAttrib();
}

void pop_projection_matrix()
{
  glPushAttrib(GL_TRANSFORM_BIT);
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glPopAttrib();
}

void print_GL_font(double x,vect yv,vect *z,const char *fmt, ...)
{
  GLuint f;
  double y;
  char text[1000];
  va_list ap;
  GLfloat modelview_matrix[16];

  y=yv.v[0];
  f=font.list_base;
  if (fmt == NULL) return;
  va_start(ap, fmt);
  vsprintf(text, fmt, ap);
  va_end(ap);

  if (z==NULL) {
    pushScreenCoordinateMatrix();
    glPushAttrib(GL_LIST_BIT | GL_CURRENT_BIT  | GL_ENABLE_BIT |
						  GL_TRANSFORM_BIT);
    glMatrixMode(GL_MODELVIEW);
    glDisable(GL_LIGHTING);
    glEnable(GL_TEXTURE_2D);
    glDisable(GL_DEPTH_TEST);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glListBase(f);
    glGetFloatv(GL_MODELVIEW_MATRIX, modelview_matrix);
    glPushMatrix();
    glLoadIdentity();
    glTranslatef(x,y,0);
    glMultMatrixf(modelview_matrix);
    glCallLists(strlen(text), GL_UNSIGNED_BYTE, text);

    glPopMatrix();

    glPopAttrib();
    pop_projection_matrix();
  } else {
    glLoadIdentity();
    glTranslatef(gl.x,gl.y,gl.z);
    glRotatef(gl.phix,1.0,0.0,0.0);
    glRotatef(gl.phiz,0.0,0.0,1.0);
    glRotatef(gl.phiy,0.0,1.0,0.0);
    glRotatef(-90.0,1.0,0.0,0.0);
    glTranslatef(gl.f*z->v[0],gl.f*z->v[1],gl.f*z->v[2]);
    glRotatef(90.0,1.0,0.0,0.0);
    glRotatef(-gl.phiy,0.0,1.0,0.0);
    glRotatef(-gl.phiz,0.0,0.0,1.0);
    glRotatef(-gl.phix,1.0,0.0,0.0);
    f=font.my_list_base;
    glRotatef(yv.v[0],0.0,0.0,1.0);
    glScalef(yv.v[1],yv.v[2],1.0);
    glListBase(f);
    glCallLists(strlen(text), GL_UNSIGNED_BYTE, text);
  }
}

/* transfer from printf format to normal */
int DoText(char *ss,constrglobal *cg,char **error)
{
  int a,b,c,d,e,i,parnum,quote,vectnum,*vecte;
  char *t,*s,nulla[2]="0";

  s=strdup(ss);
  vectnum=parnum=0;quote=(s[0]=='"');
  for (a=quote;a<(int)strlen(s);a++) {
    if ((quote) && (s[a]=='"')) break;
    if ((!quote) && (s[a]==',')) break;
    if (s[a]=='%') {
      a++;
      if (a==(int)strlen(s)) break;
      if ((s[a]=='%') || (s[a]=='"')) continue;
      if (s[a]=='v') {
	parnum+=2;vectnum++;
      }
      parnum++;
    }
  }
  if (parnum==0) {
    t=s;
    if (quote) {
      t++;
      a=0;
      while ((a<(int)strlen(t)) && (t[a]!='"')) a++;
      t[a]='\0';
    }
    cg->text=strdup(t);
    cg->textpnum=0;
  } else {
    vecte=(int *)calloc(parnum,sizeof(int));
    t=(char *)calloc(strlen(s)+20+vectnum*30,sizeof(char));
    i=0;
    for (a=quote,b=0;a<(int)strlen(s)-(s[strlen(s)-1]=='"');a++,b++) {
      if ((quote) && (s[a]=='"')) break;
      if ((!quote) && (s[a]==',')) break;
      if ((s[a]=='%') && (s[a+1]=='"')) {
	a++;
	t[b]=s[a];
      } else if ((s[a]=='%') && (s[a+1]=='v')) {
	vecte[i++]=1;vecte[i++]=1;vecte[i++]=1;
	c=a+2;
	while ((c<(int)strlen(s)) && (tolower(s[c])!='f') &&
	       (tolower(s[c])!='e') && (tolower(s[c])!='g') &&
	       (s[c]!='%')) c++;
	a+=2;
	if ((c==(int)strlen(s)) || (s[c]=='%')) c=a;
	t[b++]='(';
	for (d=0;d<3;d++) {
	  t[b++]='%';
	  if (c>a) {
	    for (e=a;e<c;e++) {
	      t[b++]=s[e];
	    }
	  } else t[b++]='g';
	  t[b++]=',';
	}
	b--;
	t[b]=')';
      } else if ((s[a]=='%') && (s[a+1]=='%')) {
	t[b++]=s[a++];
	t[b]=s[a];
      } else {
	if (s[a]=='%') vecte[i++]=0;
	t[b]=s[a];
      }
    }
    if (a<(int)strlen(s)) a++;
    if ((quote) && (s[a]==',')) a++;
    t[b]='\0';
    cg->text=strdup(t);
    free(t);
    cg->textpnum=parnum;
    cg->textparam=(RPN **) calloc(parnum,sizeof(RPN *));
    for (b=0;b<parnum;b++) {
      cg->textparam[b]=NULL;
    }
    for (b=0;b<parnum;b++) {
      c=a;
      while ((a<(int)strlen(s)) && (s[a]!=',')) a++; 
      e=0;
      if (a<(int)strlen(s)) {
	s[a]='\0';e=1;
      }
      if (strlen(&s[c])==0) {
	if (vecte[b]) {
	  RPN_decode(nulla,&cg->textparam[b++],error,0,&d,
			BG_RPN_globalvarnum,bg.rpnglobstr);
	  RPN_decode(nulla,&cg->textparam[b++],error,0,&d,
			BG_RPN_globalvarnum,bg.rpnglobstr);
	  RPN_decode(nulla,&cg->textparam[b],error,0,&d,
			BG_RPN_globalvarnum,bg.rpnglobstr);
	} else {
	  RPN_decode(nulla,&cg->textparam[b],error,0,&d,
			BG_RPN_globalvarnum,bg.rpnglobstr);
	}
      } else {
	if (vecte[b]) {
	  t=(char *)calloc(strlen(&s[c])+20,sizeof(char));
	  for (i=0;i<3;i++,b++) {
	    sprintf(t,"Comp(%s,%d)",&s[c],i);
	    if (!RPN_decode(t,&cg->textparam[b],error,0,&d,
			BG_RPN_globalvarnum,bg.rpnglobstr)) {
	      free(vecte);
	      free(t);
	      free(s);
	      return(0);
	    }
	  }
	  b--;
	  free(t);
	} else {
	  if (!RPN_decode(&s[c],&cg->textparam[b],error,0,&d,
			BG_RPN_globalvarnum,bg.rpnglobstr)) {
	    free(vecte);
	    free(s);
	    return(0);
	  }
	}
      }
      if (e) s[a++]=',';
    }
    free(vecte);
  }
  free(s);
  return(1);
}

void Plot_Font(constrglobal *cg,int mode)
{
  int a;
  double *rv,dv,x;
  vect v,*z,w,y;

  z=NULL;
  if (cg->p[1]!=NULL) {
    RPN_Evaluate(cg->p[1],body[0].params,bg.rpnglobpoint,
			&dv,NULL,NULL,0,0);
    if ((dv<0.0) && (!mode)) {
      cg->type=-11;
      return;
    } else if ((dv<0.0) && (mode)) {
      dv=-dv;
      cg->type=11;
    }
    x=max(min(dv,1.0),0.0)*gl.width;
  } else x=0.0;
  if (cg->textpnum) rv=(double *)calloc(cg->textpnum,sizeof(double));
  if (cg->p[2]!=NULL) {
    RPN_Evaluate(cg->p[2],body[0].params,bg.rpnglobpoint,
			&dv,&w,&a,0,0);
    if (a) {
      z=&w;
    } else {
      y.v[0]=max(min(dv,1.0),0.0)*gl.height;
    }
  } else y.v[0]=0.0;
  if (cg->p[3]!=NULL) {
    RPN_Evaluate(cg->p[3],body[0].params,bg.rpnglobpoint,
			NULL,&v,NULL,0,0);
  } else {
    v.v[0]=v.v[2]=0.0;v.v[1]=1.0;
  }
  glColor4f(v.v[0],v.v[1],v.v[2],1.0);
  glLoadIdentity();
  glTranslatef(0,0,0);
  if (cg->p[4]!=NULL) {
    RPN_Evaluate(cg->p[4],body[0].params,bg.rpnglobpoint,
			&dv,NULL,NULL,0,0);
  } else dv=0.0;
  glRotatef(dv,0.0,0.0,1.0);
  if (z!=NULL) y.v[0]=dv;
  if (cg->p[5]!=NULL) {
    RPN_Evaluate(cg->p[5],body[0].params,bg.rpnglobpoint,
			&dv,&v,&a,0,0);
  } else {
    dv=1.0;
    a=0;
  }
  if (a) {
    glScalef(v.v[0],v.v[1],1.0);
    if (z!=NULL) {
      y.v[1]=v.v[0];
      y.v[2]=v.v[1];
    }
  } else {
    glScalef(dv,dv,1.0);
    if (z!=NULL) y.v[1]=y.v[2]=dv;
  }
  for (a=0;a<cg->textpnum;a++) {
    RPN_Evaluate(cg->textparam[a],body[0].params,bg.rpnglobpoint,
			&rv[a],NULL,NULL,0,0);
  }
  switch (cg->textpnum) {
    case 0:
      print_GL_font(x,y,z,cg->text);
      break;
    case 1:
      print_GL_font(x,y,z,cg->text,rv[0]);
      break;
    case 2:
      print_GL_font(x,y,z,cg->text,rv[0],rv[1]);
      break;
    case 3:
      print_GL_font(x,y,z,cg->text,rv[0],rv[1],rv[2]);
      break;
    case 4:
      print_GL_font(x,y,z,cg->text,rv[0],rv[1],rv[2],rv[3]);
      break;
    case 5:
      print_GL_font(x,y,z,cg->text,rv[0],rv[1],rv[2],rv[3],rv[4]);
      break;
    case 6:
      print_GL_font(x,y,z,cg->text,rv[0],rv[1],rv[2],rv[3],rv[4],rv[5]);
      break;
    case 7:
      print_GL_font(x,y,z,cg->text,rv[0],rv[1],rv[2],rv[3],rv[4],rv[5],rv[6]);
      break;
    case 8:
      print_GL_font(x,y,z,cg->text,rv[0],rv[1],rv[2],rv[3],rv[4],rv[5],rv[6],rv[7]);
      break;
    case 9:
      print_GL_font(x,y,z,cg->text,rv[0],rv[1],rv[2],rv[3],rv[4],rv[5],rv[6],rv[7],rv[8]);
      break;
    case 10:
      print_GL_font(x,y,z,cg->text,rv[0],rv[1],rv[2],rv[3],rv[4],rv[5],rv[6],rv[7],rv[8],rv[9]);
      break;
    case 11:
      print_GL_font(x,y,z,cg->text,rv[0],rv[1],rv[2],rv[3],rv[4],rv[5],rv[6],rv[7],rv[8],rv[9],rv[10]);
      break;
    case 12:
      print_GL_font(x,y,z,cg->text,rv[0],rv[1],rv[2],rv[3],rv[4],rv[5],rv[6],rv[7],rv[8],rv[9],rv[10],rv[11]);
      break;
    case 13:
      print_GL_font(x,y,z,cg->text,rv[0],rv[1],rv[2],rv[3],rv[4],rv[5],rv[6],rv[7],rv[8],rv[9],rv[10],rv[11],rv[12]);
      break;
    case 14:
      print_GL_font(x,y,z,cg->text,rv[0],rv[1],rv[2],rv[3],rv[4],rv[5],rv[6],rv[7],rv[8],rv[9],rv[10],rv[11],rv[12],rv[13]);
      break;
    case 15:
      print_GL_font(x,y,z,cg->text,rv[0],rv[1],rv[2],rv[3],rv[4],rv[5],rv[6],rv[7],rv[8],rv[9],rv[10],rv[11],rv[12],rv[13],rv[14]);
      break;
    case 16:
      print_GL_font(x,y,z,cg->text,rv[0],rv[1],rv[2],rv[3],rv[4],rv[5],rv[6],rv[7],rv[8],rv[9],rv[10],rv[11],rv[12],rv[13],rv[14],rv[15]);
      break;
    case 17:
      print_GL_font(x,y,z,cg->text,rv[0],rv[1],rv[2],rv[3],rv[4],rv[5],rv[6],rv[7],rv[8],rv[9],rv[10],rv[11],rv[12],rv[13],rv[14],rv[15],rv[16]);
      break;
    case 18:
      print_GL_font(x,y,z,cg->text,rv[0],rv[1],rv[2],rv[3],rv[4],rv[5],rv[6],rv[7],rv[8],rv[9],rv[10],rv[11],rv[12],rv[13],rv[14],rv[15],rv[16],rv[17]);
      break;
    case 19:
      print_GL_font(x,y,z,cg->text,rv[0],rv[1],rv[2],rv[3],rv[4],rv[5],rv[6],rv[7],rv[8],rv[9],rv[10],rv[11],rv[12],rv[13],rv[14],rv[15],rv[16],rv[17],rv[18]);
      break;
  }
  if (cg->textpnum) free(rv);
}

void reset_font(void)
{
  font.font=NULL;
}
