/* Copyright (C) 2002, by Sandor Fazekas and Janos Torok */
/* Last modified: $Date: 2004/09/01 13:04:05 $ $Revision: 1.9 $ */

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include "io.h"
#include "global.h"
#include "object.h"
#include "rpn.h"
#include "font.h"
#include "object.h"

#ifdef _WIN32
#pragma warning( disable : 4090 )
#endif

/* Prototypes */
void io_setfiletype(int);
void Read_Params(FILE *,char *);

typedef struct Param_ {
  const char *name;     /* parameter name */
  int type;             /* type */
  void *data;           /* where data is to be stored */
  int is_req;           /* if it is required */
} Param;

/* types */
#define PARI    1       /* int */
#define PARD    2       /* double */
#define PARS    3       /* string */
#define LNUM    4       /* light num */
#define LPAR    5       /* light struct */
#define COLS    6       /* background */
#define PATS    7       /* tree string */

static Param param_list[] = {
  { "filename",         PARS, &io.filename,             0 },
  { "width",            PARI, &gl.width,                0 },
  { "height",           PARI, &gl.height,               0 },
  { "frameskip",        PARI, &anim.frameskip,          0 },
  { "dx",               PARD, &gl.x,                    0 },
  { "dy",               PARD, &gl.y,                    0 },
  { "dz",               PARD, &gl.z,                    0 },
  { "phix",             PARD, &gl.phix,                 0 },
  { "phiy",             PARD, &gl.phiy,                 0 },
  { "phiz",             PARD, &gl.phiz,                 0 },
  { "quadres",          PARI, &gl.quad.st,              0 },
  { "factor",           PARD, &gl.f,                    0 },
  { "bg",               COLS, &gl.bg,                   0 },
  { "adjustframedt",    PARI, &anim.adjustframedt,      0 },
  { "adjustquadres",    PARI, &anim.adjustquadres,      0 },
  { "framedt",          PARI, &anim.framedt,            0 },
  { "keyfactor",        PARD, &anim.keyf,               0 },
  { "deg",              PARI, &RPN_deg,                 0 },
  { "numchange",        PARD, &og.numchange,            0 },
  { "ttffont",          PARS, &font.font,               0 },
  { "constraint",       PATS, &mparam_cstrp,            0 },
};

#define PARAM_NUM sizeof(param_list)/sizeof(Param)

void Write_rc(char *home, int a);

int Read_rc(char *prg)
{
  FILE *in;
  char *home,*fullname;

  /* check current directory first */
  in=fopen(anim.rcfile,"r");
  if (in!=NULL) {
    mparam_cstr_start=NULL;
    mparam_cstrp=&mparam_cstr_start;
    io_setfiletype(0);    /* rc file should be ascii */
    Read_Params(in,prg);
    fclose(in);
  } else {
    home=getenv("HOME");
    if (home==NULL) home=strdup("");
    fullname=(char *) calloc(strlen(home)+strlen(anim.rcfile)+2,sizeof(char));
    sprintf(fullname,"%s/%s",home,anim.rcfile);
    in=fopen(fullname,"r");
    if (in==NULL) return(0);
    mparam_cstr_start=NULL;
    mparam_cstrp=&mparam_cstr_start;
    io_setfiletype(0);    /* rc file should be ascii */
    Read_Params(in,prg);
    fclose(in);
  }
  return(1);
}

void Write_Constraint(int mode)
{
  int a;
  FILE *out,*cin;
  char *home,*fullname,*fullname2,*fullname3;
  char s[256];
  size_t new, save;

  home=getenv("HOME");
  if (home==NULL) home=strdup("");
  fullname=(char *) calloc(strlen(home)+strlen(anim.rcfile)+2,sizeof(char));
  sprintf(fullname,"%s/%s",home,anim.rcfile);
  fullname2=(char *) calloc(strlen(home)+strlen(anim.rcfile)+3,sizeof(char));
  sprintf(fullname2,"%s/%s~",home,anim.rcfile);
  fullname3=strdup(fullname2);
  fullname3[strlen(fullname2)-1]='_';
  if (!mode) {
    remove(fullname2);
    rename(fullname,fullname2);
    cin=fopen(fullname2,"r");
    out=fopen(fullname,"w");
    while (!feof(cin)) {
      fgets(s,256,cin);
      if (strncmp(s,"constraint =",strlen("constraint ="))==0) break;
      if (!feof(cin)) {
	fputs(s,out);
      }
    }
    fclose(cin);
    fclose(out);
  }
  out=fopen(fullname,"a");
  if (named_constraint_start!=NULL) {
    named_constraintp=&named_constraint_start;
    while (named_constraintp!=NULL) {
      fprintf(out,"constraint = %s\n",(*named_constraintp)->s);
      if ((*named_constraintp)->next==NULL) named_constraintp=NULL;
      else named_constraintp=&(*named_constraintp)->next;
    }
  }
  for (a=og.fhstart;a!=og.fhend;a=(a+1)%CH_Max) {
    fprintf(out,"constraint = %s\n",og.fnamehist[a]);
  }
  fclose(out);

  /* Check length */
  cin=fopen(fullname,"r");
  if (cin) {
    fseek(cin,0,SEEK_END);
    new=ftell(cin);
    fclose(cin);
  } else new=0;
  cin=fopen(fullname3,"r");
  if (cin) {
    fseek(cin,0,SEEK_END);
    save=ftell(cin);
    fclose(cin);
  } else save=0;
  if (save<new) {
    cin=fopen(fullname,"r");
    out=fopen(fullname3,"w");
    while (!feof(cin)) {
      a=fgetc(cin);
      if (!feof(cin)) fputc(a,out);
    }
    fclose(cin);
    fclose(out);
  }
  free(fullname3);
  free(fullname2);
  free(fullname);
}

void Write_rc_cur(void) {

  char *tmp;
  tmp = getenv("PWD");
  Write_rc(tmp, 0);

}

void Write_rc_usr(void) {

  char *tmp;
  tmp = getenv("HOME");
  Write_rc(tmp, 1);

}

/* @param a: if a=0, filename is written into .mozirc */
void Write_rc(char *home, int a)
{
  fprintf(stderr,"%s\n", io.filename);
  FILE *out;
  double *dp;
  char *fullname;

  if (home==NULL) home=strdup("");
  fullname=(char *) calloc(strlen(home)+strlen(anim.rcfile)+2,sizeof(char));
  sprintf(fullname,"%s/%s",home,anim.rcfile);
  fprintf(stderr,"target: %s/%s\n",home,anim.rcfile);
  out=fopen(fullname,"w");
  free(fullname);
  if (out==NULL) return;
  for (;a<PARAM_NUM;a++) {
    if (param_list[a].type!=PATS) fprintf(out,"%s = ",param_list[a].name);
    switch (param_list[a].type) {
      case PARS:
        fprintf(out,"%s\n",*((char **)param_list[a].data));
        break;
      case PARI:
        fprintf(out,"%d\n",*((int *)param_list[a].data));
        break;
      case PARD:
        fprintf(out,"%le\n",*((double *)param_list[a].data));
        break;
      case COLS:
        dp=(double *)(param_list[a].data);
        fprintf(out,"%le %le %le %le\n",*dp,*(dp+1),*(dp+2),*(dp+3));
        break;
    }
  }
  fclose(out);
  Write_Constraint(1);
  fprintf(stderr,".mozirc is written\n");
}

int Which_Param(char *p)
{
  int a;

  for (a=0;a<PARAM_NUM;a++) {
    if (strcmp(p,param_list[a].name)==0) {
      return(a);
    }
  }
  return(-1);
}

void Read_Params(FILE *in,char *prg)
{
  int ppos;
  int a;
  char *p,*pp;
  double *dp;

  while (!io.eof(in)) {
    if (!io.getline(in)) {
      p=io.buff;
      /* remove '\n at the end */
      if (p[strlen(p)-1]=='\n') p[strlen(p)-1]='\0';
      /* remove whitespaces */
      while (isspace(*p)) p++;
      if (*p!='#') {      /* comment */
	pp=p;
	while (!isspace(*pp)) pp++;
	*pp='\0';
	ppos=Which_Param(p);
	if (ppos>=0) {
	  p=pp+1;
	  while ((isspace(*p)) || (*p=='=')) p++;
	  /* now set value */
	  switch (param_list[ppos].type) {
	    case PARI:    /* int */
	      *((int *)(param_list[ppos].data))=atoi(p);
	      param_list[ppos].is_req=0;
	      break;
	    case PARD:    /* double */
	      *((double *)(param_list[ppos].data))=atof(p);
	      param_list[ppos].is_req=0;
	      break;
	    case PARS:    /* string no static is assumed */
	      if (*((char **)(param_list[ppos].data))!=NULL) {
		free(*((char **)(param_list[ppos].data)));
	      }
	      *((char **)(param_list[ppos].data))=strdup(p);
	      param_list[ppos].is_req=0;
	      break;
	    case COLS:
	      dp=(double *)(param_list[ppos].data);
	      sscanf(p,"%le %le %le %le",dp,dp+1,dp+2,dp+3);
	      param_list[ppos].is_req=0;
	      break;
	    case PATS:
	      *mparam_cstrp=(constrstr *) malloc(sizeof(constrstr));
	      (*mparam_cstrp)->s=strdup(p);
	      (*mparam_cstrp)->next=NULL;
	      mparam_cstrp=&((*mparam_cstrp)->next);
	      break;
	    default:
	      fprintf(stderr,"%s error: unknown data type\n",prg);
	      exit(1);
	  }
	} else {
	  fprintf(stderr,"%s warning: unknown parameter:\n",prg);
	  fprintf(stderr,"> %s\n",p);
	}
      }
    }
  }
  for (a=0;a<PARAM_NUM;a++) {
    if (param_list[a].is_req) {
      fprintf(stderr,"%s error: parameter %s was not initialized\n",
          prg,param_list[a].name);
      exit(1);
    }
  }
}
