#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "global.h"
#include "io.h"
#include "anim.h"
#include "light.h"
#include "rpn.h"
#include "body.h"

#ifdef _WIN32
/*'conversion' conversion from 'type1' to 'type2', possible loss of data*/
#pragma warning(disable:4244)
#endif

int ilog(double);
void perr(char *, ...);
void io_setfiletype(int);
int io_detfiletype(void);
int move_bp(char *,int);
int Constraint(int a,int b,int mode);
void SetVar(void);
void SetTrans(void);
void CheckFont(int a);
void Snapshot(GtkWidget *,gpointer);

/* These settings are valid for all objects */
void Set_General(void)
{
  glEnable(GL_COLOR_MATERIAL);
  glColor4f(lb.color[0],lb.color[1],lb.color[2],lb.color[3]);
  if (lb.tex) {
    glBindTexture(GL_TEXTURE_2D,gl.texture.texture[lb.tex-1]);
    glEnable(GL_TEXTURE_2D);
  } else {
    glDisable(GL_TEXTURE_2D);
  }

  if (lb.blend) {
    glEnable(GL_BLEND);
  } else glDisable(GL_BLEND);

  glLoadIdentity();
  glTranslatef(gl.x,gl.y,gl.z);
  glRotatef(gl.phix,1.0,0.0,0.0);
  glRotatef(gl.phiz,0.0,0.0,1.0);
  glRotatef(gl.phiy,0.0,1.0,0.0);
  glRotatef(-90.0,1.0,0.0,0.0);
  glTranslatef(gl.f*lb.x, gl.f*lb.y,gl.f*lb.z);
  glRotatef(lb.phi*RP,lb.rx,lb.ry,lb.rz);
}

void Plot_Body(int a,int mode)
{
  int b,rv;
  double r,q;

  if ((a+anim.constrreread) && (body[a].obj!=bg.obj_num) &&
      (!bg.param[body[a].obj].show)) return;
  if ((mode) && (!body[a].blend)) return;
  if ((!mode) && (body[a].blend)) return;
  if ((!anim.donotdoublebody) && (anim.constrreread==1)) {
    if (!restb[a].show) return;
    lb.type=restb[a].type;
    lb.blend=restb[a].blend;
    lb.tex=restb[a].tex;
    lb.color[0]=restb[a].color[0];
    lb.color[1]=restb[a].color[1];
    lb.color[2]=restb[a].color[2];
    lb.color[3]=restb[a].color[3];
    lb.x=restb[a].x;
    lb.y=restb[a].y;
    lb.z=restb[a].z;
    lb.rx=restb[a].rx;
    lb.ry=restb[a].ry;
    lb.rz=restb[a].rz;
    lb.phi=restb[a].phi;
    lb.sx=restb[a].sx;
    lb.sy=restb[a].sy;
    lb.sz=restb[a].sz;
  } else {
    lb.type=body[a].type;
    lb.blend=body[a].blend;
    lb.tex=body[a].tex;
    lb.color[0]=body[a].color[0];
    lb.color[1]=body[a].color[1];
    lb.color[2]=body[a].color[2];
    lb.color[3]=body[a].color[3];
    lb.x=body[a].x;
    lb.y=body[a].y;
    lb.z=body[a].z;
    lb.rx=body[a].rx;
    lb.ry=body[a].ry;
    lb.rz=body[a].rz;
    lb.phi=body[a].phi;
    lb.sx=body[a].sx;
    lb.sy=body[a].sy;
    lb.sz=body[a].sz;
    for (b=0;b<CEF_PARNUM*CEF_SUBPARNUM;b++) {
      lb.params[b]=body[a].params[b];
    }
    if (a==0) {
      rv=bg.param[body[a].obj].show;
      for (b=0;b<bg.cgnum;b++) {
	if (Constraint(a,b,mode)) {
	  if (!anim.donotdoublebody) {
	    restb[a].type=lb.type;
	    restb[a].show=0;
	  }
	  rv=0;
	}
      }
      if (!rv) return;
    } else {
      for (b=0;b<bg.cgnum;b++) {
	if (Constraint(a,b,mode)) {
	  if (!anim.donotdoublebody) {
	    restb[a].type=lb.type;
	    restb[a].show=0;
	  }
	  return;
	}
      }
    }
    if (!anim.donotdoublebody) {
      restb[a].show=1;
      restb[a].type=lb.type;
      restb[a].blend=lb.blend;
      restb[a].tex=lb.tex;
      restb[a].color[0]=lb.color[0];
      restb[a].color[1]=lb.color[1];
      restb[a].color[2]=lb.color[2];
      restb[a].color[3]=lb.color[3];
      restb[a].x=lb.x;
      restb[a].y=lb.y;
      restb[a].z=lb.z;
      restb[a].rx=lb.rx;
      restb[a].ry=lb.ry;
      restb[a].rz=lb.rz;
      restb[a].phi=lb.phi;
      restb[a].sx=lb.sx;
      restb[a].sy=lb.sy;
      restb[a].sz=lb.sz;
    }
  }
  Set_General();
  r=lb.sx*gl.f*0.5;
  switch (lb.type) {
    case BT_Wall:
      q=lb.sy*0.5*gl.f;
      glBegin(GL_QUADS);
      glNormal3f(0.0,0.0,1.0);
      glTexCoord2f(0.0, 1.0);
      glVertex3f(-r, q, 0.0);
      glTexCoord2f(0.0, 0.0);
      glVertex3f(-r,-q, 0.0);
      glTexCoord2f(1.0, 0.0);
      glVertex3f( r,-q, 0.0);
      glTexCoord2f(1.0, 1.0);
      glVertex3f( r, q, 0.0);
      glEnd();
      break;
    case BT_Ellipsoid:
      glScalef(lb.sx*gl.f,lb.sy*gl.f,lb.sz*gl.f);
      r=0.5;
    case BT_DGBall:
    case BT_STBall:
    case BT_BallCC:
    case BT_Ball:
      gluSphere(gl.quad.q, r, gl.quad.sl,gl.quad.st);
      break;
    case BT_Force:
    case BT_DTEdge:
    case BT_Cylinder:
      gluCylinder(gl.quad.q,fabs(r),fabs(r),fabs(lb.sz)*gl.f,
	  gl.quad.sl,gl.quad.st);
      break;
    case BT_Disk:
      gluDisk(gl.quad.q, 0.0, r, gl.quad.sl,gl.quad.st);
      break;
    case BT_Arrow:
      q=lb.sz*gl.f;
      r*=2.0;
      gluCylinder(gl.quad.q,q,0.0,r,gl.quad.sl,gl.quad.st);
      glTranslatef(0.0,0.0,r*0.7);
      gluCylinder(gl.quad.q,2.0*q,0.0,r*0.3,gl.quad.sl,gl.quad.st);
      break;
    case BT_Cone:
      gluCylinder(gl.quad.q,r,lb.sy*gl.f*0.5,lb.sz*gl.f,gl.quad.sl,gl.quad.st);
      break;
    case BT_Triangle:
      r*=2.0;
      glBegin(GL_QUADS);
      glNormal3f(0.0,0.0,1.0);
      glTexCoord2f(0.0, 1.0);
      glVertex3f(0.0,0.0, 0.0);
      glTexCoord2f(0.0, 0.0);
      glVertex3f(0.0,0.0, 0.0);
      glTexCoord2f(1.0, 0.0);
      glVertex3f( r,0.0, 0.0);
      glTexCoord2f(1.0, 1.0);
      glVertex3f(lb.sy*gl.f,lb.sz*gl.f, 0.0);
      glEnd();
      break;
    case BT_Triangle2:
      r*=2.0;
      glBegin(GL_QUADS);
      glNormal3f(0.0,0.0,1.0);
      glTexCoord2f(0.0,1.0);
      glVertex3f(0.0,0.0,0.0);
      glTexCoord2f(0.0, 0.0);
      glVertex3f(0.0,0.0,0.0);
      glTexCoord2f(1.0,0.0);
      glVertex3f(lb.rx*gl.f,lb.ry*gl.f,lb.rz*gl.f);
      glTexCoord2f(1.0, 1.0);
      glVertex3f(r,lb.sy*gl.f,lb.sz*gl.f);
      glEnd();
      break;
    case BT_NULL:
      /* do nothing */
      break;
  }
}

void Plot_Objects(void)
{
  int a;

  glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);
  glColorMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE);
  if (!bg.save_new) {
    bg.new_num=body_alloc;
  }
  SetVar();
  SetTrans();
  CheckFont(0);
  for (a=0;a<bg.body_num;a++) Plot_Body(a,0);
  for (a=bg.new_num;a<body_alloc;a++) Plot_Body(a,0);
  glFlush();
  if (ll.moveit) {      /* if light is moved then plot the sun(s) */
    GLfloat suncolor[4]={1.0,1.0,0.0,1.0};

    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE,suncolor);
    glEnable(GL_COLOR_MATERIAL);
    glColor4f(1.0,1.0,0.0,1.0);
    glDisable(GL_TEXTURE_2D);
    for (a=0;a<MaxLight;a++) {
      if (gl.light[a].on) {
        glLoadIdentity();
        glTranslatef(gl.x,gl.y,gl.z);
        glTranslatef(gl.light[a].param[0][0],gl.light[a].param[0][1],
                        gl.light[a].param[0][2]);
        gluSphere(gl.quad.q, 0.1, gl.quad.sl,gl.quad.st);
      }
    }
  }
  for (a=0;a<bg.body_num;a++) Plot_Body(a,1);
  for (a=bg.new_num;a<body_alloc;a++) Plot_Body(a,1);
  bg.gbeg[bg.obj_num]=bg.new_num;
  bg.gnum[bg.obj_num]=body_alloc-bg.new_num;
  CheckFont(1);
  glFlush();
  if (anim.constrreread>0) anim.constrreread=1;
  if (anim.save_png) Snapshot(NULL,(gpointer)0);
}

void Print_Warning(int a,int v)
{
  static int was[2];
  int i;
#define W_AllocP 0
#define W_AllocN 1

  switch (a) {
    case 0:
      if (was[0]) break;
      fprintf(stderr,"Mozi warning: More objects (in line %d) than ",v);
      fprintf(stderr,"body_alloc(=%d). Use -b flag.\n",body_alloc);
      was[0]=1;
      break;
    case 1:
      if (was[1]) break;
      fprintf(stderr,"Mozi warning: In first line more objects (%d) than ",v);
      fprintf(stderr,"body_alloc(=%d).\nUse -b flag.\n",body_alloc);
      was[1]=1;
      break;
    case -1:
      for (i=0;i<2;i++) was[i]=0;
  }
}

void Set_FPos(void)
{
  int a;
  for (a=0;a<bg.body_num;a++) {
    body[a].params[(CEF_PARNUM-1)*CEF_SUBPARNUM]=(double)animation.framepc;
    body[a].params[(CEF_PARNUM-1)*CEF_SUBPARNUM+2]=
		(double)animation.framepc*animation.framerepeat+
			animation.framerepeatcounter;
  }
}

void Read_Objects(int pos,int det)
{
  int a,b,c,i0,i1,i2,n,d,i,ll,bs,on,ondv,obj,foursix,idx;
  int times,run;
  int typ,tex,num,dummyint,N[3],ix[3],objtype[OBJMAX];
  int cg_i,dxm,dxp,dym,dyp,dzm,dzp,base,pp;
  double x,y,z,rx,ry,rz,phi,time;
  double low[OBJMAX][3],hi[OBJMAX][3];
  double *rho,dx,dxx,dyy,dzz,mass;
  char *p;

  if ((!det) && (anim.coarse_grain)) {
    for (b=0;b<OBJMAX;b++) {
      for (c=0;c<3;c++) {
	low[b][c]=1e100;hi[b][c]=-1e100;
      }
    }
    objtype[b]=0;
  }
  anim.objectread=1;
  if (anim.constrreread>0) anim.constrreread=2;
  animation.loadpc=-1;
  for (b=0;b<bg.cgnum;b++) {
    if (cg[b].type == CG_Load) {
      Constraint(0,b,1);
    }
  }
  obj=0;
  ll=bg.line_length;
  pp=pos;
  bg.body_num=bg.wall_num;
  if (animation.loadpc>=0) times=2;
  else times=1;
  run=0;
  do {
    if (bg.line_length<=0) {
      io.seek(io.in,animation.framepos[pp],SEEK_SET);
      if ((bg.input_type%2==0) && (bg.input_type>1)) {
	ll=animation.framepos[pp+1]-animation.framepos[pp];
	a=io.read(io.in,ll);
	if (a!=ll) {
	  fprintf(stderr,"Mozi warning: in data file type %d at",bg.input_type);
	  fprintf(stderr," pos %d short line (%d/%d)\n",pp,a,bg.line_length);
	  return;
	}
      } else {
	/* text input */
	io.getline(io.in);
      }
      io.scan("%d",&a);time=(double)a;
    } else {
      /* we have fixed line length */
      io.seek(io.in,animation.framepos[0]+pp*bg.line_length,SEEK_SET);
      a=io.read(io.in,bg.line_length);
      if (a!=bg.line_length) {
	fprintf(stderr,"Mozi warning: in data file type %d at",bg.input_type);
	fprintf(stderr," pos %d short line (%d/%d)\n",pp,a,bg.line_length);
	return;
      }
    }
    foursix=0;
    switch (bg.input_type) {
      case 6:
	foursix=6;
      case 4:
	i=sizeof(int);d=sizeof(double);
	p=io.buff;
	if (run==0) {
	  obj=bg.body_num-1;
	  time=*((double *)p);
	}
	p+=d;
	/* object type is byte, end if it is zero */
	while ((*p) && (p<ll+io.buff)) {
	  typ=*((int *)p);p+=i;
	  if (foursix) {
	    tex=*((int *)p);p+=i;
	  } else  {
	    tex=*p++;
	  }
	  num=*((int *)p);p+=i;
	  bs=bg.body_num;
	  bg.body_num+=max(det,num);
	  obj++;
	  if (run>0) bg.param[obj].show=1;
	  if (bg.gbeg!=NULL) {
	    bg.gbeg[obj]=bs;
	    bg.gnum[obj]=bg.body_num-bs;
	  }
	  objtype[obj]=typ&BP_Mask;
	  if (bg.body_num>body_alloc) {
	    Print_Warning(W_AllocP,pp);
	    bg.body_num=body_alloc;
	  }
	  for (a=bs;a<bg.body_num;a++) {
	    if (a>body_alloc) break;
	    if (foursix) {
	      body[a].tex=(tex&31);
	      body[a].color[0]=((tex>>8)&255)/255.0;
	      body[a].color[1]=((tex>>16)&255)/255.0;
	      body[a].color[2]=((tex>>24)&255)/255.0;
	      body[a].color[3]=0.5;
	    } else {
	      body[a].tex=(tex&15)+1;
	    }
	    body[a].type=typ&BP_Mask;
	    body[a].par=typ;
	    /* determine object from different source only at reset */
	    if (!anim.groupbytex) body[a].obj=obj;
	    if (num==0) break;
	    /* get size of body. It's compulsory */
	    if (body[a].type!=BT_NULL) {
	      body[a].sx=*((double *)p);p+=d;
	    }
	    if ((body[a].type==BT_Cylinder) || (body[a].type==BT_Force) ||
		(body[a].type==BT_DTEdge) || (body[a].type==BT_Arrow)) {
	      body[a].sz=*((double *)p);p+=d;
	    }
	    if (body[a].type==BT_BallCC) {
	      body[a].params[5*CEF_SUBPARNUM]=(double)(*((int *)p));p+=i;
	      body[a].sy=body[a].params[5*CEF_SUBPARNUM];
	    }
	    if (body[a].type==BT_Wall) {
	      body[a].sy=*((double *)p);p+=d;
	    }
	    if ((body[a].type==BT_Force) || (body[a].type==BT_DTEdge)) {
	      body[a].ind1=*((int *)p);p+=i;
	      body[a].ind2=*((int *)p);p+=i;
	    }
	    if ((body[a].type==BT_Ellipsoid) || (body[a].type==BT_Cone) ||
		(body[a].type==BT_Triangle) || (body[a].type==BT_Triangle2)) {
	      body[a].sy=*((double *)p);p+=d;
	      body[a].sz=*((double *)p);p+=d;
	    }
	    /* position */
	    if (typ&BP_Pos) {
	      body[a].x=*((double *)p);p+=d;
	      body[a].y=*((double *)p);p+=d;
	      body[a].z=*((double *)p);p+=d;
	      if ((!det) && (anim.coarse_grain)) {
		if (body[a].x<low[obj][0]) low[obj][0]=body[a].x;
		if (body[a].y<low[obj][1]) low[obj][1]=body[a].y;
		if (body[a].z<low[obj][2]) low[obj][2]=body[a].z;
		if (body[a].x>hi[obj][0]) hi[obj][0]=body[a].x;
		if (body[a].y>hi[obj][1]) hi[obj][1]=body[a].y;
		if (body[a].z>hi[obj][2]) hi[obj][2]=body[a].z;
	      }
	    } else body[a].x=body[a].y=body[a].z=0.0;
	    /* orientation */
	    if (typ&BP_Orient) {
	      body[a].rx=*((double *)p);p+=d;
	      body[a].ry=*((double *)p);p+=d;
	      body[a].rz=*((double *)p);p+=d;
	      body[a].phi=*((double *)p);p+=d;
	    } else {
	      body[a].rx=body[a].ry=body[a].phi=0.0;
	      body[a].rz=1.0;
	    }
	    /* other parameters */
	    on=5*CEF_SUBPARNUM;
	    if (typ&BP_Coord) {
	      bs=*((int *)p);
	      body[a].params[on]=(double)bs;
	      p+=i;
	    }
	    on+=CEF_SUBPARNUM;
	    if (typ&BP_Velo) {
	      body[a].params[on]=*((double *)p);p+=d;
	      body[a].params[on+1]=*((double *)p);p+=d;
	      body[a].params[on+2]=*((double *)p);p+=d;
	    }
	    on+=CEF_SUBPARNUM;
	    if (typ&BP_Force) {
	      body[a].params[on]=*((double *)p);p+=d;
	      body[a].params[on+1]=*((double *)p);p+=d;
	      body[a].params[on+2]=*((double *)p);p+=d;
	    }
	    on+=CEF_SUBPARNUM;
	    if (typ&BP_AngVelo) {
	      body[a].params[on]=*((double *)p);p+=d;
	      body[a].params[on+1]=*((double *)p);p+=d;
	      body[a].params[on+2]=*((double *)p);p+=d;
	    }
	    on+=CEF_SUBPARNUM;
	    if (typ&BP_Torque) {
	      body[a].params[on]=*((double *)p);p+=d;
	      body[a].params[on+1]=*((double *)p);p+=d;
	      body[a].params[on+2]=*((double *)p);p+=d;
	    }
	    on+=CEF_SUBPARNUM;
	    if (typ&BP_TrVelo) {
	      body[a].params[on]=*((double *)p);p+=d;
	      body[a].params[on+1]=*((double *)p);p+=d;
	      body[a].params[on+2]=*((double *)p);p+=d;
	    }
	    on+=CEF_SUBPARNUM;
	    if (typ&BP_TrForce) {
	      body[a].params[on]=*((double *)p);p+=d;
	      body[a].params[on+1]=*((double *)p);p+=d;
	      body[a].params[on+2]=*((double *)p);p+=d;
	    }
	    on+=CEF_SUBPARNUM;
	    if (typ&BP_TrAngVelo) {
	      body[a].params[on]=*((double *)p);p+=d;
	      body[a].params[on+1]=*((double *)p);p+=d;
	      body[a].params[on+2]=*((double *)p);p+=d;
	    }
	    if (typ&BP_Handle) {
	      bs=*((int *)p);
	      body[a].params[2]=(double)bs;
	      p+=i;
	    }
	    on+=CEF_SUBPARNUM;
	    if (typ&BP_LifeTime) {
	      body[a].params[on]=*((double *)p);p+=d;
	    }
	    on+=CEF_SUBPARNUM;
	    if (typ&BP_MatPar) {
	      body[a].params[on]=*((double *)p);p+=d;
	      body[a].params[on+1]=*((double *)p);p+=d;
	      body[a].params[on+2]=*((double *)p);p+=d;
	    }
	    on+=CEF_SUBPARNUM;
	    if (typ&BP_TrCoord) {
	      body[a].params[on]=*((double *)p);p+=d;
	    }
	    on+=CEF_SUBPARNUM;
	    if (typ&BP_TrTorque) {
	      body[a].params[on]=*((double *)p);p+=d;
	      body[a].params[on+1]=*((double *)p);p+=d;
	      body[a].params[on+2]=*((double *)p);p+=d;
	    }
	    if (tex&16) body[a].appearance ^= BA_Texture2D;
	    if (tex&32) body[a].appearance |= BA_Blend;
	    body[a].blend=(tex&32)/32;
	  }
	}
	if ((det) && (!anim.groupbytex)) bg.obj_num=obj+1;
	break;
      case 3:
	perr("data file type 3 is not supported in mozi\n");
	break;
      case 2:
	perr("data file type 2 is not any more supported in mozi\n");
	i=sizeof(int);d=sizeof(double);
	p=io.buff;
	time=*((double *)p);p+=d;
	obj=bg.body_num-1;
	/* object type is byte, end if it is zero */
	while ((*p) && (p<bg.line_length+io.buff)) {
	  typ=*p++;
	  objtype[obj]=typ&BP_Mask;
	  tex=*p++;
	  num=*((int *)p);p+=i;
	  obj++;
	  for (a=bg.body_num;a<bg.body_num+num;a++) {
	    if (a>body_alloc) break;
	    body[a].type=typ;
	    /* determine object from different source only at reset */
	    if (!anim.groupbytex) body[a].obj=obj;
	    body[a].x=*((double *)p);p+=d;
	    body[a].y=*((double *)p);p+=d;
	    body[a].z=*((double *)p);p+=d;
	    body[a].rx=*((double *)p);p+=d;
	    body[a].ry=*((double *)p);p+=d;
	    body[a].rz=*((double *)p);p+=d;
	    body[a].phi=*((double *)p);p+=d;
	    body[a].sx=*((double *)p);p+=d;
	    if (body[a].type==BT_Cylinder) {
	      body[a].sz=*((double *)p);p+=d;
	    }
	    if (body[a].type==BT_BallCC) p+=i;
	    if (body[a].type==BT_Wall) {
	      body[a].sy=*((double *)p);p+=d;
	    }
	    if ((!det) && (anim.coarse_grain)) {
	      if (body[a].x<low[obj][0]) low[obj][0]=body[a].x;
	      if (body[a].y<low[obj][1]) low[obj][1]=body[a].y;
	      if (body[a].z<low[obj][2]) low[obj][2]=body[a].z;
	      if (body[a].x>hi[obj][0]) hi[obj][0]=body[a].x;
	      if (body[a].y>hi[obj][1]) hi[obj][1]=body[a].y;
	      if (body[a].z>hi[obj][2]) hi[obj][2]=body[a].z;
	    }
	    body[a].tex=(tex&15)+1;
	    body[a].appearance=BA_Color+BA_Texture2D;
	    if (tex&16) body[a].appearance ^= BA_Texture2D;
	    if (tex&32) body[a].appearance |= BA_Blend;
	    body[a].blend=(tex&32)/32;
	  }
	  bg.body_num+=num;
	}
	if ((det) && (!anim.groupbytex)) bg.obj_num=obj+1;
	if (bg.body_num>body_alloc) {
	  Print_Warning(W_AllocP,pp);
	  bg.body_num=body_alloc;
	}
	break;
      case 1:
	while ((io.scan("%d",&typ)==1)  && (bg.body_num<body_alloc)) {
	  io.scan("%d",&tex); io.scan("%d",&num);
	  obj--;
	  for (a=bg.body_num;a<bg.body_num+num;a++) {
	    if (a>body_alloc) break;
	    body[a].type=typ;
	    objtype[obj]=typ&BP_Mask;
	    if (det) body[a].obj=obj;
	    io.scan("%le",&body[a].x);
	    io.scan("%le",&body[a].y);
	    io.scan("%le",&body[a].z);
	    io.scan("%le",&body[a].rx);
	    io.scan("%le",&body[a].ry);
	    io.scan("%le",&body[a].rz);
	    io.scan("%le",&body[a].phi);
	    io.scan("%le",&body[a].sx);
	    if ((!det) && (anim.coarse_grain)) {
	      if (body[a].x<low[obj][0]) low[obj][0]=body[a].x;
	      if (body[a].y<low[obj][1]) low[obj][1]=body[a].y;
	      if (body[a].z<low[obj][2]) low[obj][2]=body[a].z;
	      if (body[a].x>hi[obj][0]) hi[obj][0]=body[a].x;
	      if (body[a].y>hi[obj][1]) hi[obj][1]=body[a].y;
	      if (body[a].z>hi[obj][2]) hi[obj][2]=body[a].z;
	    }
	    if (body[a].type==BT_Cylinder) io.scan("%le",&body[a].sz);
	    if (body[a].type==BT_BallCC) io.scan("%d",&dummyint);
	    if (body[a].type==BT_Wall) io.scan("%le",&body[a].sy);
	    body[a].tex=(tex&15)+1;
	    body[a].appearance=BA_Color+BA_Texture2D;
	    if (tex&16) body[a].appearance ^= BA_Texture2D;
	    if (tex&32) body[a].appearance |= BA_Blend;
	    body[a].blend=(tex&32)/32;
	  }
	  bg.body_num+=num;
	}
	if (bg.body_num>body_alloc) {
	  Print_Warning(W_AllocP,pp);
	  bg.body_num=body_alloc;
	}
	break;
      case 0:
      default:
	io.scan("%d",&n); /* read the number of balls */
	bg.body_num+=n;
	if (bg.body_num>body_alloc) {
	  Print_Warning(W_AllocP,pp);
	  bg.body_num=body_alloc;
	}
	for (a=bg.body_num-n;a<bg.body_num;a++) {
	  io.scan("%le",&x);io.scan("%le",&y);io.scan("%le",&z);
	  io.scan("%le",&rx);io.scan("%le",&ry);io.scan("%le",&rz);
	  io.scan("%le",&phi);
	  body[a].type=BT_Ball;
	  body[a].sx=0.001;
	  body[a].x=x;
	  body[a].y=y;
	  body[a].z=z;
	  if ((!det) && (anim.coarse_grain)) {
	    if (body[a].x<low[obj][0]) low[obj][0]=body[a].x;
	    if (body[a].y<low[obj][1]) low[obj][1]=body[a].y;
	    if (body[a].z<low[obj][2]) low[obj][2]=body[a].z;
	    if (body[a].x>hi[obj][0]) hi[obj][0]=body[a].x;
	    if (body[a].y>hi[obj][1]) hi[obj][1]=body[a].y;
	    if (body[a].z>hi[obj][2]) hi[obj][2]=body[a].z;
	  }
	  body[a].rx=rx;
	  body[a].ry=ry;
	  body[a].rz=rz;
	  body[a].phi=phi;
	  if (!anim.groupbytex) body[a].obj=bg.wall_num;
	  objtype[bg.wall_num]=1;
	}
	if (!anim.groupbytex) bg.obj_num=bg.wall_num+1;
	break;
    }
    pp=animation.loadpc%animation.maxframe;
    times--;
    run++;
  } while (times>0);
  if (bg.obj_num) {
    if ((det) && (anim.coarse_grain)) {
      bg.cg_pos=bg.obj_num;
      c=0;
      for (b=0;b<bg.obj_num;b++) {
	if ((objtype[b]==BT_Ball) || (objtype[b]==BT_BallCC)) c++;
      }
      for (a=bg.body_num;a<bg.body_num+c;a++) {
	body[a].type=BT_STBall;
	body[a].sx=anim.cg_length;
	body[a].par=BP_Pos+BP_Velo+BP_Dvx+BP_Dvy+BP_Dvz;
      }
      bg.body_num+=c;
      bg.obj_num+=c;
    }
    if ((!det) && (anim.coarse_grain)) {
      dx=anim.cg_length;
      /* coarse graining */
      N[0]=N[1]=N[2]=1;
      cg_i=0;
      for (b=0;b<bg.obj_num;b++) {
	/* determine grid size */
	if (bg.obj_num-bg.cg_pos<=cg_i) break;
	if ((objtype[b]!=2) && (objtype[b]!=5)) continue;
	cg_i++;
	for (c=0;c<3;c++) {
	  if ((hi[b][c]-low[b][c])/dx>N[c]) {
	    N[c]=(int)((hi[b][c]-low[b][c])/dx);
	  }
	}
      }
      rho=(double *)malloc(sizeof(double)*N[0]*N[1]*N[2]);
      cg_i=0;
      for (b=0;b<bg.obj_num;b++) {
	/* calculate lengths again */
	if (bg.obj_num-bg.cg_pos<=cg_i) break;
	if ((objtype[b]!=2) && (objtype[b]!=5)) continue;
	cg_i++;
	N[0]=N[1]=N[2]=1;
	for (c=0;c<3;c++) {
	  if ((hi[b][c]-low[b][c])/dx>N[c]) {
	    N[c]=(int)((hi[b][c]-low[b][c])/dx);
	  }
	}
	if (bg.body_num+N[0]*N[1]*N[2]>body_alloc) {
	  Print_Warning(W_AllocP,pos);
	  break;
	}
	base=bg.body_num;
	for (i0=0;i0<N[0];i0++) {
	  for (i1=0;i1<N[1];i1++) {
	    for (i2=0;i2<N[2];i2++) {
	      idx=i0*N[1]*N[2]+i1*N[2]+i2;
	      body[base+idx].par=9;
	      body[base+idx].obj=obj+cg_i;
	      body[base+idx].sx=dx;
	      body[base+idx].sy=dx;
	      body[base+idx].sz=dx;
	      body[base+idx].tex=0;
	      body[base+idx].color[0]=1.0;
	      body[base+idx].color[1]=1.0;
	      body[base+idx].color[2]=0.0;
	      body[base+idx].color[3]=0.5;
	      body[base+idx].x=(i0+0.5)*dx*(N[0]>1)+low[b][0];
	      body[base+idx].y=(i1+0.5)*dx*(N[1]>1)+low[b][1];
	      body[base+idx].z=(i2+0.5)*dx*(N[2]>1)+low[b][2];
	      on=BP_Force_I*CEF_SUBPARNUM;
	      body[base+idx].params[on]=0.0;
	      body[base+idx].params[on+1]=0.0;
	      body[base+idx].params[on+2]=0.0;
	      on=BP_Dvx_I*CEF_SUBPARNUM;
	      body[base+idx].params[on]=0.0;
	      body[base+idx].params[on+1]=0.0;
	      body[base+idx].params[on+2]=0.0;
	      on+=CEF_SUBPARNUM;
	      body[base+idx].params[on]=0.0;
	      body[base+idx].params[on+1]=0.0;
	      body[base+idx].params[on+2]=0.0;
	      on+=CEF_SUBPARNUM;
	      body[base+idx].params[on]=0.0;
	      body[base+idx].params[on+1]=0.0;
	      body[base+idx].params[on+2]=0.0;
	      rho[idx]=0.0;
	    }
	  }
	}
	for (a=bg.gbeg[b];a<bg.gbeg[b]+bg.gnum[b];a++) {
	  ix[0]=(int)((body[a].x-low[b][0])/dx);
	  ix[1]=(int)((body[a].y-low[b][1])/dx);
	  ix[2]=(int)((body[a].z-low[b][2])/dx);
	  on=BP_Force_I*CEF_SUBPARNUM;
	  ondv=BP_MatPar_I*CEF_SUBPARNUM+1;
	  for (i0=max(0,ix[0]-5);i0<min(N[0],ix[0]+6);i0++) {
	    for (i1=max(0,ix[1]-5);i1<min(N[1],ix[1]+6);i1++) {
	      for (i2=max(0,ix[2]-5);i2<min(N[2],ix[2]+6);i2++) {
		phi=exp(-(pow(i0*dx+low[b][0]-body[a].x,2.0)+
		    pow(i1*dx+low[b][1]-body[a].y,2.0)+
		    pow(i2*dx+low[b][2]-body[a].z,2.0))/dx/dx)/(M_PI*dx*dx);
		idx=i0*N[1]*N[2]+i1*N[2]+i2;
		/* there seems to be some inconsistency between 2d and 3d
		 * CD snapshots in the mass definition if cleared can be
		 * used here */
		if (typ&BP_MatPar) {
		  mass=body[a].params[ondv];
		} else {
		  mass=body[a].sx*body[a].sx*body[a].sx/6.0*M_PI;
		}
/*		mass=1.0;*/
		rho[idx]+=mass*phi;
		for (c=0;c<3;c++) {
		  body[base+idx].params[on+c]+=mass*phi*body[a].params[on+c];
		}
	      }
	    }
	  }
	}
	for (i0=0;i0<N[0];i0++) {
	  for (i1=0;i1<N[1];i1++) {
	    for (i2=0;i2<N[2];i2++) {
	      idx=i0*N[1]*N[2]+i1*N[2]+i2;
	      for (c=0;c<3;c++) {
		body[base+idx].params[on+c]/=rho[idx];
	      }
	    }
	  }
	}
	for (i0=0;i0<N[0];i0++) {
	  dxm=(i0>0        ? -N[1]*N[2] : 0);
	  dxp=(i0<(N[0]-1) ?  N[1]*N[2] : 0);
	  dxx=((i0==0) || (i0==(N[0]-1)) ? 1.0:2.0);
	  for (i1=0;i1<N[1];i1++) {
	    dym=(i1>0        ? -N[2] : 0);
	    dyp=(i1<(N[1]-1) ?  N[2] : 0);
	    dyy=((i1==0) || (i1==(N[1]-1)) ? 1.0:2.0);
	    for (i2=0;i2<N[2];i2++) {
	      dzm=(i2>0        ? -1 : 0);
	      dzp=(i2<(N[2]-1) ?  1 : 0);
	      dzz=((i2==0) || (i2==(N[2]-1)) ? 1.0:2.0);
	      idx=i0*N[1]*N[2]+i1*N[2]+i2;
	      ondv=BP_Dvx_I*CEF_SUBPARNUM;
	      on=BP_Force_I*CEF_SUBPARNUM;
	      for (c=0;c<3;c++) {
		body[base+idx].params[ondv]=(body[base+idx+dxp].params[on]-
		    body[base+idx+dxm].params[on])/dx/dxx;
		body[base+idx].params[ondv+1]=(body[base+idx+dyp].params[on]-
		    body[base+idx+dym].params[on])/dx/dyy;
		body[base+idx].params[ondv+2]=(body[base+idx+dzp].params[on]-
		    body[base+idx+dzm].params[on])/dx/dzz;
		ondv+=CEF_SUBPARNUM;on++;
	      }
	    }
	  }
	}
	bg.gbeg[bg.cg_pos+cg_i-1]=bg.body_num;
	bg.gnum[bg.cg_pos+cg_i-1]=N[0]*N[1]*N[2];
	bg.body_num+=N[0]*N[1]*N[2];
      }
      free(rho);
    }
    bs=0;
    for (a=0;a<bg.body_num;a++) {
      /* time is set here */
      body[a].params[(CEF_PARNUM-1)*CEF_SUBPARNUM]=(double)pos;
      body[a].params[(CEF_PARNUM-1)*CEF_SUBPARNUM+1]=time;
      body[a].params[(CEF_PARNUM-1)*CEF_SUBPARNUM+2]=
	(double)pos*animation.framerepeat+animation.framerepeatcounter;
      if (anim.groupbytex) {
        while (body[a].type!=bg.param[bs].type) bs++;
        body[a].obj=bs;
      }
      body[a].params[0]=(double)a;
      body[a].params[1]=(double)body[a].obj+1.0;
      on=CEF_SUBPARNUM;
      body[a].params[on]=body[a].sx;
      body[a].params[on+1]=body[a].sy;
      body[a].params[on+2]=body[a].sz;
      body[a].params[on+3]=(double)body[a].ind1;
      body[a].params[on+4]=(double)body[a].ind2;
      on+=CEF_SUBPARNUM;
      body[a].params[on]=body[a].x;
      body[a].params[on+1]=body[a].y;
      body[a].params[on+2]=body[a].z;
      on+=CEF_SUBPARNUM;
      body[a].params[on]=body[a].rx;
      body[a].params[on+1]=body[a].ry;
      body[a].params[on+2]=body[a].rz;
      body[a].params[on+3]=body[a].phi;
      on+=CEF_SUBPARNUM;
      if (bg.input_type<5) {
        body[a].color[0]=0.8f;
        body[a].color[1]=0.6f;
        body[a].color[2]=1.0f;
        body[a].color[3]=0.4f;
      }
      body[a].params[on]=body[a].color[0];
      body[a].params[on+1]=body[a].color[1];
      body[a].params[on+2]=body[a].color[2];
      body[a].params[on+3]=body[a].color[3];
    }
  }
}

void Not_Fixed_Linelength()
{
  long i,d,d3,v,llmax,typ,num,lpos,first;
  int bases[BT_N_BODY];
  long *buff,sumbodynum;

  i=sizeof(int);d=sizeof(double);d3=3*d;
  bases[BT_EOL]=0;bases[BT_Wall]=d+d;bases[BT_Ball]=d;
  bases[BT_Cylinder]=d+d;bases[BT_Disk]=d;bases[BT_BallCC]=d+i;
  bases[BT_Force]=d+d+i+i;bases[BT_STBall]=d;bases[BT_DGBall]=d;
  bases[BT_DTEdge]=d+d+i+i;bases[BT_Ellipsoid]=d+d+d;
  bases[BT_Arrow]=d+d;bases[BT_Cone]=d+d+d;bases[BT_Triangle]=d+d+d;
  bases[BT_Triangle2]=d+d+d;bases[BT_NULL]=0;
  llmax=0;lpos=io.tell(io.in);
  sumbodynum=0;first=1;
  do {
    if (llmax<io.tell(io.in)-lpos) llmax=io.tell(io.in)-lpos;
    animation.framepos[animation.maxframe++]=io.tell(io.in);
    if (animation.maxframe==animation.framepossize) {
      /* buffer is full */
      buff=animation.framepos;
      animation.framepos=(long *)
                      calloc(animation.framepossize*2,sizeof(long));
      memcpy(animation.framepos,buff,animation.framepossize*sizeof(long));
      free(buff);
      animation.framepossize*=2;
    }
    io.seek(io.in,d,SEEK_CUR);
    do {
      io.read(io.in,i);
      typ=*((int *)io.buff);
      if (typ) {
        if (bg.input_type<5) {
          io.seek(io.in,1,SEEK_CUR);
        } else {
          io.seek(io.in,i,SEEK_CUR);
        }
        v=io.read(io.in,i);
        num=*((int *)io.buff);
	if ((v==0) || (num<0)) {
	  animation.maxframe--;
	  if (animation.maxframe==0) animation.maxframe=1;
	  return;
	}
	sumbodynum+=num;
        v=bases[typ&BP_Mask];
        if (typ&BP_Pos) v+=d3;
        if (typ&BP_Orient) v+=d3+d;
        if (typ&BP_Coord) v+=i;
        if (typ&BP_Velo) v+=d3;
        if (typ&BP_Force) v+=d3;
        if (typ&BP_AngVelo) v+=d3;
        if (typ&BP_Torque) v+=d3;
        if (typ&BP_TrVelo) v+=d3;
        if (typ&BP_TrForce) v+=d3;
        if (typ&BP_TrAngVelo) v+=d3;
        if (typ&BP_TrTorque) v+=d3;
        if (typ&BP_Handle) v+=i;
        if (typ&BP_LifeTime) v+=d;
        if (typ&BP_MatPar) v+=d3;
        if (typ&BP_TrCoord) v+=d;
        io.seek(io.in,v*num,SEEK_CUR);
      } else if (first) {
	if (body_alloc<sumbodynum) Print_Warning(W_AllocN,sumbodynum);
	first=0;
      }
    } while ((!io.eof(io.in)) && (typ));
  } while (!io.eof(io.in));
  if (llmax<io.tell(io.in)-lpos) llmax=io.tell(io.in)-lpos;
  animation.maxframe--;
  if (animation.maxframe==0) animation.maxframe=1;
}

void reset_objects()
{
  int a,tex,typ,p,num,size_int,size_dble,goon,comment;
  long pos,*buff;
  char ch;

  bg.line_length=-1;
  if (!io_detfiletype()) perr("input file %s does not exist\n",io.filename);
  io.in=io.open(io.filename,"rb");
  if (io.in==NULL) perr("Input file %s not found\n",io.filename);

  body=(BODY *)calloc(body_alloc+1,sizeof(BODY));
  if (!anim.donotdoublebody) {
    restb=(RestBODY *)calloc(body_alloc+1,sizeof(RestBODY));
  }
  for (a=0;a<body_alloc;a++) {
    body[a].type=BT_Ball;
    body[a].x=0.0;
    body[a].y=0.0;
    body[a].z=0.0;
    body[a].rx=0.0;
    body[a].ry=0.0;
    body[a].rz=1.0;
    body[a].phi=0.0;
    body[a].sx=0.001;
    body[a].sy=0.0;
    body[a].sz=0.0;
    body[a].appearance=BA_Color+BA_Texture2D;
    body[a].blend=0;
    body[a].tex=2;
    body[a].par=96;
  }
  bg.wall_num=0;

  /* detect frame positions */
  animation.framepossize=1000;
  animation.framepos=(long *) calloc(animation.framepossize,sizeof(long));
  animation.maxframe=0;
  animation.framepc=0;
  bg.save_new=0;
  bg.input_type=0;goon=1;
  comment=1;
  do {
    pos=io.tell(io.in);
    io.getline(io.in);
    ch=io.buff[0];
    if (ch!='#') comment=0;
    if ((!io.eof(io.in)) && (comment==0)) {
      /* skip comments */
      animation.framepos[animation.maxframe++]=pos;
      if (animation.maxframe==animation.framepossize) {
        /* buffer is full */
        buff=animation.framepos;
        animation.framepos=(long *)
                        calloc(animation.framepossize*2,sizeof(long));
        memcpy(animation.framepos,buff,animation.framepossize*sizeof(long));
        free(buff);
        animation.framepossize*=2;
      }
    }
    if (comment) {       /* special */
      if (strlen(io.buff)==2) { /* A single # character in a line */
        if (bg.line_length>0) {
          /* We have the position information in the header so we quit */
          animation.framepos[0]=pos=io.tell(io.in);
          animation.framepossize=
                animation.maxframe=(io.tell(io.in)-pos)/bg.line_length;
        } else {
          Not_Fixed_Linelength();
        }
        goon=0;
        break;
      }
      if (strncmp(io.buff+1,"object",6)==0) {   /* indeed */
        p=1+strlen("object");
        sscanf(io.buff+p,"%d %d %d",&typ,&tex,&num);
        p+=move_bp(io.buff+p,3);
        for (a=bg.wall_num;a<bg.wall_num+num;a++) {
          body[a].type=typ;
          sscanf(io.buff+p,"%lf %lf %lf %lf %lf %lf %lf %lf",
              &body[a].x,&body[a].y,&body[a].z,&body[a].rx,
              &body[a].ry,&body[a].rz,&body[a].phi,&body[a].sx);
          p+=move_bp(io.buff+p,8);
          if ((body[a].type==BT_Cylinder) || (body[a].type==BT_BallCC)) {
            sscanf(io.buff+p,"%lf",&body[a].sz);
            p+=move_bp(io.buff+p,1);
          }
          if (body[a].type==BT_Wall) {
            sscanf(io.buff+p,"%lf",&body[a].sy);
          }
          body[a].tex=(tex&15)+1;
          body[a].appearance=BA_Color+BA_Texture2D;
          if (tex&16) body[a].appearance ^= BA_Texture2D;
          if (tex&32) body[a].appearance |= BA_Blend;
	  body[a].blend=(tex&32)/32;
        }
        bg.wall_num+=num;
      } else if (strncmp(io.buff+1,"type",4)==0) {      /* type is different */
        p=1+strlen("type");
        sscanf(io.buff+p,"%d",&bg.input_type);
        if ((bg.input_type==3) || (bg.input_type==2)) {
          perr("data file type 2, 3 is not supported in mozi\n");
        }
        if (bg.input_type==4) {
          /* read size of int/dble, line length */
          p+=move_bp(io.buff+p,1);
          sscanf(io.buff+p,"%d %d %d",&size_int,&size_dble,&bg.line_length);
          if ((size_int!=sizeof(int)) || (size_dble!=sizeof(double))) {
            perr("Size of int or double does not match!\n");
          }
          if (io.linelen<bg.line_length) {
            /* we reallocate io.buff */
            free(io.buff);
            io.linelen=bg.line_length+1;
            io.buff=(char *)calloc(io.linelen,sizeof(char));
          }
        }
      }
    }
  } while ((!io.eof(io.in)) && (goon));
  animation.maxframe--;
  if (animation.maxframe==0) animation.maxframe=1;
  bg.body_num=bg.wall_num;
  io.rewind(io.in);
  if (animation.marker2<0) {
    animation.marker2=animation.maxframe;
  }
  bg.obj_num=0;

}

void reset_BodyDisplaySettings()
{
  int a,b,typ,tex;

  anim.save_png=0;
  anim.constrreread=0;
  bg.gbeg=NULL;
  animation.loadpc=-1;
  Read_Objects(0,1);    /* read objects we have */
  if (anim.groupbytex) {
    /* group by texture and type */
    bg.obj_num=0;typ=tex=0;
    for (a=0;a<bg.body_num;a++) {
      if ((body[a].type!=typ) || (body[a].tex!=tex)) {
	bg.obj_num++;
	typ=body[a].type;tex=body[a].tex;
      }
    }
    /* allocate memory */
    bg.param=(bodydisplaystr *) calloc(bg.obj_num*3+1,sizeof(bodydisplaystr));
    bg.gbeg=(int *)calloc(bg.obj_num*3+1,sizeof(int));
    bg.gnum=(int *)calloc(bg.obj_num*3+1,sizeof(int));
    b=typ=tex=0;
    for (a=0;a<bg.body_num;a++) {
      if ((body[a].type!=typ) || (body[a].tex!=tex)) {
	bg.gbeg[b]=a;
	bg.gnum[b]=0;
	typ=body[a].type;tex=body[a].tex;
	if (anim.init_show) bg.param[b].show=1;
	else bg.param[b].show=0;
	bg.param[b].type=typ;
	bg.param[b].par=body[a].par;
	bg.param[b].blend=body[a].appearance&BA_Blend;
	bg.param[b].texture=((body[a].appearance&BA_Texture2D) ? tex : 0);
	bg.param[b].color[0]=0.8f;
	bg.param[b].color[1]=0.6f;
	bg.param[b].color[2]=1.0f;
	bg.param[b].color[3]=0.4f;
	b++;
      }
      bg.gnum[b-1]++;
      body[a].obj=b-1;
    }
  } else {
    bg.param=(bodydisplaystr *) calloc(bg.obj_num+1,sizeof(bodydisplaystr));
    bg.gbeg=(int *)calloc(bg.obj_num+1,sizeof(int));
    bg.gnum=(int *)calloc(bg.obj_num+1,sizeof(int));
    for (a=0,b=0;a<bg.wall_num;a++,b++) {
      bg.gbeg[b]=a;
      bg.gnum[b]=1;
      typ=body[a].type;tex=body[a].tex;
      if (anim.init_show) bg.param[b].show=1;
      else bg.param[b].show=0;
      bg.param[b].type=typ;
      bg.param[b].par=body[a].par;
      bg.param[b].blend=body[a].appearance&BA_Blend;
      bg.param[b].texture=((body[a].appearance&BA_Texture2D) ? tex : 0);
      bg.param[b].color[0]=0.8f;
      bg.param[b].color[1]=0.6f;
      bg.param[b].color[2]=1.0f;
      bg.param[b].color[3]=0.4f;
      body[a].obj=b;
    }
    for (;b<bg.obj_num;b++) {
      for (;a<bg.body_num;a++) {
	if (body[a].obj==b) break;
      }
      bg.gbeg[b]=a;
      if (b>0) bg.gnum[b-1]=bg.gbeg[b]-bg.gbeg[b-1];
      typ=body[a].type;tex=body[a].tex;
      if (anim.init_show) bg.param[b].show=1;
      else bg.param[b].show=0;
      bg.param[b].type=typ;
      bg.param[b].par=body[a].par;
      bg.param[b].blend=body[a].appearance&BA_Blend;
      bg.param[b].texture=((body[a].appearance&BA_Texture2D) ? tex : 0);
      bg.param[b].color[0]=0.8f;
      bg.param[b].color[1]=0.6f;
      bg.param[b].color[2]=1.0f;
      bg.param[b].color[3]=0.4f;
    }
    bg.param[bg.obj_num].show=1;
    bg.param[bg.obj_num].type=2;
    bg.param[bg.obj_num].par=96;
    bg.param[bg.obj_num].blend=0;
    bg.param[bg.obj_num].texture=3;
    bg.param[bg.obj_num].color[0]=0.9f;
    bg.param[bg.obj_num].color[1]=0.7f;
    bg.param[bg.obj_num].color[2]=0.2f;
    bg.param[bg.obj_num].color[3]=1.0f;
  }
  if (bg.obj_num>1) bg.gnum[bg.obj_num-1]=bg.body_num-bg.gnum[bg.obj_num-2];
  else bg.gnum[bg.obj_num-1]=bg.body_num;
  bg.new_num=body_alloc;
  bg.gbeg[bg.obj_num]=bg.new_num;
  bg.gnum[bg.obj_num]=body_alloc-bg.new_num;
  Print_Warning(-1,0);
  Read_Objects(0,0);    /* read objects we have */
}

/* EOF */
